#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "levelset.h"
#include "vectors.h"

void read_level_set(const char *filename, LevelSet *ls)
{
   FILE *fp;
   int i, numentries;

   fp=fopen(filename, "rt");
   if(!fp) return;
   fscanf(fp, "%d %d %d", &ls->nx, &ls->ny, &ls->nz); // grid dimensions
   fscanf(fp, "%f", &ls->dx); // grid spacing
   fscanf(fp, "%f %f %f", &ls->x0, &ls->y0, &ls->z0); // grid corner
   ls->x1=ls->x0+ls->dx*(ls->nx-1); // the opposite grid corner
   ls->y1=ls->y0+ls->dx*(ls->ny-1);
   ls->z1=ls->z0+ls->dx*(ls->nz-1);
   numentries=ls->nx*ls->ny*ls->nz;
   ls->phi=(float*)malloc(numentries*sizeof(float));
   for(i=0; i<numentries; ++i)
      fscanf(fp, "%f", ls->phi+i);
   fclose(fp);
}

void write_level_set(const char *filename, const LevelSet *ls)
{
   FILE *fp;
   int i;

   fp=fopen(filename, "wt");
   if(!fp) return;
   fprintf(fp, "%d %d %d\n", ls->nx, ls->ny, ls->nz);
   fprintf(fp, "%.7g\n", ls->dx);
   fprintf(fp, "%.7g %.7g %.7g\n", ls->x0, ls->y0, ls->z0);
   for(i=0; i<ls->nx*ls->ny*ls->nz; ++i){
      fprintf(fp, "%.7g", ls->phi[i]);
      if(i%7==6) fprintf(fp, "\n");
      else fprintf(fp, " ");
   }
   fprintf(fp, "\n");
   fclose(fp);
}

void make_level_set_of_block(LevelSet *ls, float xrange, float yrange, float zrange, float dx, float margin)
{
   // fill in this code to create a new level set for the block that goes from -xrange/2 to xrange/2,
   // -yrange/2 to yrange/2, -zrange/2 to zrange/2. Use a grid spacing of approximately dx, and allow for
   // the given margin around the block (so the grid actually covers -xrange/2-margin to xrange/2+margin, etc.)
   // You can be approximate with the margin too.
}

void make_level_set_of_sphere(LevelSet *ls, float radius, float dx, float margin)
{
   // fill in this code to create a new level set for the sphere of given radius centered at the origin.
   // Use a grid spacing of approximately dx, and allow forr the given margin around the sphere (so the grid
   // approximately covers -radius-margin to radius+margin).
}

float ls_phi(const LevelSet *ls, float x, float y, float z)
{
   int ongrid=1;
   float offx=0, offy=0, offz=0; // how far off the grid a point is, if it's off the grid
   float gx, gy, gz; // x,y,z transformed to grid coordinates
   float gi, gj, gk; // floor of gx, gy, gz
   float fx, fy, fz; // fractional part of gx, gy, gz
   int i, j, k; // integer version of gi, gj, gk
   float phi;

   // figure out where to look in grid for x
   gx=(x-ls->x0)/ls->dx;
   if(gx<0){
      ongrid=0; offx=ls->x0-x; fx=0; i=0;
   }else if(gx>=ls->nx-1){
      ongrid=0; offx=ls->x1-x; fx=1; i=ls->nx-2;
   }else{
      gi=floor(gx);
      fx=gx-gi;
      i=(int)gi;
   }
   // figure out where to look in grid for y
   gy=(y-ls->y0)/ls->dx;
   if(gy<0){
      ongrid=0; offy=ls->y0-y; fy=0; j=0;
   }else if(gy>=ls->ny-1){
      ongrid=0; offy=ls->y1-y; fy=1; j=ls->ny-2;
   }else{
      gj=floor(gy);
      fy=gy-gj;
      j=(int)gj;
   }
   // figure out where to look in grid for z
   gz=(z-ls->z0)/ls->dx;
   if(gz<0){
      ongrid=0; offz=ls->z0-z; fz=0; k=0;
   }else if(gz>=ls->nz-1){
      ongrid=0; offz=ls->z1-z; fz=1; k=ls->nz-2;
   }else{
      gk=floor(gz);
      fz=gz-gk;
      k=(int)gk;
   }

   #define PHI(i,j,k) (ls->phi[(k)+ls->nz*((j)+ls->ny*(i))])
   phi=(1-fx)*( (1-fy)*( (1-fz)*PHI(i,j,k)     + fz*PHI(i,j,k+1) )
               +    fy*( (1-fz)*PHI(i,j+1,k)   + fz*PHI(i,j+1,k+1) ))
          +fx*( (1-fy)*( (1-fz)*PHI(i+1,j,k)   + fz*PHI(i+1,j,k+1) )
               +    fy*( (1-fz)*PHI(i+1,j+1,k) + fz*PHI(i+1,j+1,k+1) ));
   #undef PHI
   if(ongrid)
      return phi;
   else{
      if(phi<0) return -sqrt(phi*phi+offx*offx+offy*offy+offz*offz);
      else return sqrt(phi*phi+offx*offx+offy*offy+offz*offz);
   }
}

void ls_normal(const LevelSet *ls, float x, float y, float z, float normal[3])
{
   // not the most efficient way, but ok for the moment
   normal[0]=ls_phi(ls, x+ls->dx, y, z)-ls_phi(ls, x-ls->dx, y, z);
   normal[1]=ls_phi(ls, x, y+ls->dx, z)-ls_phi(ls, x, y-ls->dx, z);
   normal[2]=ls_phi(ls, x, y, z+ls->dx)-ls_phi(ls, x, y, z-ls->dx);
   normalize_v(normal);
}

