#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "rigidbody.h"
#include "vectors.h"

// globals for rendering
int fullscreen_mode=0;
int width=800, height=600;
int saved_width, saved_height;
int oldmousex=0, oldmousey=0;
enum {DONOTHING, ROTATE, ZOOM, MOVETARGET} mousemode=DONOTHING;
float camdist=6;
const float NEARCLIP_FACTOR=0.1, FARCLIP_FACTOR=10;
float nearclip, farclip;
float target[3]={0, 1, 0};
float view_target_timer=0;
int view_ground=1, view_buddha=1;
float old_time=0;
const float fovy=45;
float heading=M_PI/3, pitch=0;
float campos[3];
int file_count=0;

// globals for data
int frame=0;
const char *directory=0;
int num_rb=0;
RigidBody *rb=0;

// prototypes
void read_frame(int new_frame);
void read_geometry();
void handle_display();
void handle_reshape(int w, int h);
void handle_keypress(unsigned char key, int x, int y);
void handle_click(int button, int state, int x, int y);
void handle_drag(int x, int y);
void handle_idle();
void toggle_fullscreen();
void compute_camera_position();
void display_ground();
void display_mesh(const TriMesh *mesh);
void display_rigid_bodies();
void display_target();
void save_screen(const char *filename);
void init_lighting();
void init_material(float r, float g, float b);

// definitions
int main(int argc, char *argv[])
{
   // initialize GLUT stuff 
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
   glutInitWindowSize(width,height);
   saved_width=width;
   saved_height=height;
   glutCreateWindow("hw2_viewer");
   glutDisplayFunc(handle_display);
   glutReshapeFunc(handle_reshape);
   glutKeyboardFunc(handle_keypress);
   glutMouseFunc(handle_click);
   glutMotionFunc(handle_drag);
   glutIdleFunc(handle_idle);

   // initialize OpenGL stuff
   glEnable(GL_DEPTH_TEST);
   glClearColor(0, 0, 0, 0);
   glClearDepth(1);
   glPixelStorei(GL_PACK_ALIGNMENT, 1);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   nearclip=NEARCLIP_FACTOR*camdist;
   farclip=FARCLIP_FACTOR*camdist;
   init_lighting();
   glEnable (GL_LIGHTING);

   if(argc<2){
      fprintf(stderr, "Expecting <rigid_body_directory> as argument\n");
      return 1;
   }
   directory=argv[1];

   // initialize first frame
   read_frame(0);
   read_geometry();

   // let's go
   old_time=clock()/(float)CLOCKS_PER_SEC;
   glutPostRedisplay();
   glutMainLoop();
   return 0;
}

void init_lighting()
{
   {
      GLfloat ambient[4] = {.3, .3, .3, 1};
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
   }
   {
      GLfloat color[4] = {.7, .7, .7, 1};
      glLightfv(GL_LIGHT0, GL_DIFFUSE, color);
      glLightfv(GL_LIGHT0, GL_SPECULAR, color);
      glEnable(GL_LIGHT0);
   }
   {
      GLfloat color[4] = {.4, .4, .4, 1};
      glLightfv(GL_LIGHT1, GL_DIFFUSE, color);
      glLightfv (GL_LIGHT1, GL_SPECULAR, color);
      glEnable (GL_LIGHT1);
   }
   {
      GLfloat color[4] = {.2, .2, .2, 1};
      glLightfv (GL_LIGHT2, GL_DIFFUSE, color);
      glLightfv (GL_LIGHT2, GL_SPECULAR, color);
      glEnable (GL_LIGHT2);
   }
}

void init_material(float r, float g, float b)
{
   GLfloat f_ambient[4]={0}, f_diffuse[4]={0}, f_specular[4]={0};

   f_ambient[0]=0.1*r;    f_ambient[1]=0.1*g;    f_ambient[2]=0.1+0.2*b;
   f_diffuse[0]=0.6*r;    f_diffuse[1]=0.6*g;    f_diffuse[2]=0.4*b;
   f_specular[0]=0.3;     f_specular[1]=0.3;     f_specular[2]=0.2;

   glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, f_ambient);
   glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, f_diffuse);
   glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, f_specular);
   glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, 64);
}

void read_frame(int new_frame)
{
   char *file_name;
   FILE *fp;
   int i;

   if(new_frame<0) return;
   file_name=malloc(strlen(directory)+30);
   sprintf(file_name, "%s/rigidbodies%04d", directory, new_frame);
   fp=fopen(file_name, "rt");
   if(fp){
      frame=new_frame;
      fscanf(fp, "%d", &num_rb);
      if(rb==0)
         rb=(RigidBody*)malloc(num_rb*sizeof(RigidBody));
      for(i=0; i<num_rb; ++i)
         read_rigid_body_parameters(fp, rb+i);
      fclose(fp);
   }
   free(file_name);
}

void read_geometry()
{
   char *file_name=malloc(strlen(directory)+30);
   int i;

   for(i=0; i<num_rb; ++i){
      rb[i].phi=0;
      rb[i].mesh=(TriMesh*)malloc(sizeof(TriMesh));
      sprintf(file_name, "%s/object%02d.mesh", directory, i);
      read_tri_mesh(file_name, rb[i].mesh);
      compute_vertex_normals(rb[i].mesh);
   }
   free(file_name);
}

void handle_display()
{
   glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

   glMatrixMode (GL_MODELVIEW);
   glLoadIdentity();
   gluLookAt(campos[0], campos[1], campos[2], target[0], target[1], target[2], 0, 1, 0);

   if (view_ground) display_ground();
   display_rigid_bodies();
   if(view_target_timer>0) display_target();

   glutSwapBuffers();
}

void handle_reshape(int w, int h)
{
   width=w;
   height=h;
   glViewport(0, 0, (GLsizei)width, (GLsizei)height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(fovy, width/(float)height, nearclip, farclip);
   glutPostRedisplay();
}

void handle_keypress(unsigned char key, int x, int y)
{
   switch(key){
   case '.':
      read_frame(frame+1);
      break;
   case ',':
      read_frame(frame-1);
      break;
   case 'f':
      toggle_fullscreen();
      break;
   case 'g':
      view_ground=!view_ground;
      break;
   case 'q':
      exit(0);
   case 's':
      {
         char file_name[30];
         sprintf(file_name, "hw1_%04d.ppm", file_count);
         save_screen(file_name);
         ++file_count;
      }
      break;
   default:
      return;
   }
   glutPostRedisplay();
}

void handle_click(int button, int state, int x, int y)
{
   int mods=glutGetModifiers();
   if(state==GLUT_UP) mousemode = DONOTHING;
   else if(mods&GLUT_ACTIVE_SHIFT){
      view_target_timer=1;
      mousemode=MOVETARGET;
   }else{
      if(button==GLUT_RIGHT_BUTTON || (button==GLUT_LEFT_BUTTON && mods&GLUT_ACTIVE_CTRL))
         mousemode=ZOOM;
      else if(button==GLUT_LEFT_BUTTON)
         mousemode=ROTATE;
   }
   oldmousex=x;
   oldmousey=y;
}

void handle_drag(int x, int y)
{
   switch(mousemode){
   case ROTATE:
      heading+=0.007*(oldmousex-x);
      if(heading<-M_PI) heading+=2*M_PI;
      else if(heading>M_PI) heading-=2*M_PI;
      pitch+=0.007*(oldmousey-y);
      if(pitch<-.48*M_PI) pitch=-.48*M_PI;
      else if(pitch>.48*M_PI) pitch=.48*M_PI;
      break;
   case ZOOM:
      camdist*=pow(1.01, y-oldmousey);
      nearclip=NEARCLIP_FACTOR*camdist;
      farclip=FARCLIP_FACTOR*camdist;
      break;
   case MOVETARGET:
      target[0]-=(0.002*camdist)*cos(heading)*(oldmousex-x);
      target[1]-=(0.002*camdist)*(oldmousey-y);
      target[2]-=(0.002*camdist)*sin(heading)*(oldmousex-x);
      break;
   case DONOTHING:
      break;
   }
   oldmousex=x;
   oldmousey=y;
   compute_camera_position();
   glutPostRedisplay();
}

void handle_idle()
{
   float new_time=clock()/(float)CLOCKS_PER_SEC;
   float delta_time=new_time-old_time;
   int target_turned_off=0;
   if(delta_time<0) delta_time=0; // this can happen due to roll-over
   old_time=new_time;
   if(view_target_timer>0){
      view_target_timer-=delta_time;
      if(view_target_timer<=0){
         target_turned_off=1;
         view_target_timer=0;
      }
   }
   if(target_turned_off)
      glutPostRedisplay();
}

void toggle_fullscreen()
{
   if(fullscreen_mode){
      fullscreen_mode=0;
      width=saved_width;
      height=saved_height;
      glutReshapeWindow(saved_width, saved_height);
   }else{
      fullscreen_mode=1;
      saved_width=width;
      saved_height=height;
      glutFullScreen();
   }
}

void compute_camera_position()
{
   float ch=cos(heading), sh=sin(heading), cp=cos(pitch), sp=sin(pitch);
   campos[0]=target[0]-camdist*sh*cp;
   campos[1]=target[1]-camdist*sp;
   campos[2]=target[2]-camdist*ch*cp;
}

void display_ground()
{
   init_material(0.5,0.5,0.5);
   glBegin(GL_QUADS);
   glVertex3f(-3, 0, -3);
   glVertex3f(3, 0, -3);
   glVertex3f(3, 0, 3);
   glVertex3f(-3, 0, 3);
   glEnd();
}

void display_mesh(const TriMesh *mesh)
{
   int i;
   glBegin(GL_TRIANGLES);
   for(i=0; i<mesh->nt; ++i){
      glNormal3fv(mesh->normals+3*mesh->tri[3*i]);
      glVertex3fv(mesh->x+3*mesh->tri[3*i]);
      glNormal3fv(mesh->normals+3*mesh->tri[3*i+1]);
      glVertex3fv(mesh->x+3*mesh->tri[3*i+1]);
      glNormal3fv(mesh->normals+3*mesh->tri[3*i+2]);
      glVertex3fv(mesh->x+3*mesh->tri[3*i+2]);
   }
   glEnd();
}

void display_rigid_bodies()
{
   int i;
   float R[3][3], M[16]={0};
   
   M[15]=1;
   for(i=0; i<num_rb; ++i){
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glTranslatef(rb[i].x[0], rb[i].x[1], rb[i].x[2]);
      q_to_R(rb[i].q, R);
      // convert from row-major 3x3 to column-major 4x4 format
      M[0]=R[0][0];
      M[1]=R[1][0];
      M[2]=R[2][0];
      M[4]=R[0][1];
      M[5]=R[1][1];
      M[6]=R[2][1];
      M[8]=R[0][2];
      M[9]=R[1][2];
      M[10]=R[2][2];
      glMultMatrixf(M);
      init_material(1-i/(float)num_rb, i/(float)num_rb, 0.45*(i%3)+0.1);
      display_mesh(rb[i].mesh);
      glPopMatrix();
   }
}

void display_target()
{
   float logcd=log(camdist)/log(20.);
   float smallsize=pow(20.0, floor(logcd-.5));
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glTranslatef(target[0], target[1], target[2]);
   glColor3f(1, 1, 1);
   glutWireCube(smallsize);
   glutWireCube(20*smallsize);
   glPopMatrix();
}

void save_screen(const char *filename)
{
   FILE *fp;
   GLubyte *image_buffer;

   // get the pixels
   image_buffer=malloc(3*width*height*sizeof(GLubyte));
   glReadBuffer(GL_FRONT);
   glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, image_buffer);
  
   // write them to the file
   fp=fopen(filename, "wb");
   if(fp){
      int i;
      fprintf(fp, "P6\n%d %d 255\n", width, height);
      for(i=1; i<=height; ++i){
         fwrite(image_buffer+3*width*(height-i), 1, 3*width, fp);
      }
   }
   
   // clean up
   free(image_buffer);
}

